#!/usr/bin/ruby
# encoding: utf-8

class AntibotAviraScanner
	
	def initialize(drvscreen, params, logdir=nil)
		@drivescreen = drvscreen 
		@params = params
		@logdir = logdir
		@qparams = params.collect { |x| "'#{x}'" }  
		@infections = Hash.new
		@return_codes = Hash.new
		@killed = false
		@partcount = -1
		@logdb = nil
		@tstamp = Time.now.to_i
		@debug = AntibotMisc.boot_options["debug"]
		@avira_version = nil
		prepare_logdb
	end
	attr_reader :infections, :killed, :return_codes, :tstamp
	attr_accessor :logbd
	
	def kill_scan
		system("killall -9 scancl")
		@killed = true
	end
	
	def prepare_logdb
		return true unless @logdb.nil?
		return true if @logdir.nil? && @debug == false
		@logdb = SQLite3::Database.new(@logdir + "/fulllog.sqlite") unless @logdir.nil?
		@logdb = SQLite3::Database.new("/tmp/fulllog.sqlite") if @logdir.nil?
		begin
			@logdb.execute("CREATE TABLE single_infection " + 
				" (id INTEGER PRIMARY KEY ASC, scantime INTEGER(12), uuid VARCHAR(80), fpath VARCHAR(200), bakup VARCHAR(200), "  + 
				"  infection VARCHAR(120), scancl VARCHAR(120), sha1pre CHAR(40), sha1post CHAR(40), retvalue INT(6), " + 
				" fileout VARCHAR(120), scanout VARCHAR(800), action CHAR(20), retval INTEGER(5) ); ")
		rescue
		end
		# FIXME! Create indices!
	end
	
	def prepare_scan(pgbar=nil, label=nil)
		puts "Preparing scan:"
		now = Time.new.strftime("%Y%m%d-%H%M%S") 
		system("blkid | grep -v squashfs > '#{@logdir}/#{now}-laufwerke.txt' ") unless @logdir.nil?
		@partcount = 0
		@drivescreen.dump_marked
		( @drivescreen.int_drives + @drivescreen.ext_drives ).each { |d| 
			d.partitions.each { |p|
				if @drivescreen.saved_states[p.uuid] == true
					puts "Marked for scan: " + p.uuid
					p.filecount(pgbar)
					@partcount += 1
				end
			}
		}
		unless pgbar.nil?
			pgbar.text = @partcount.to_s + " Partitionen zu scannen"
			pgbar.text = "Virenscan abgebrochen" if @killed == true
		end
	end
	
	def check_avira(pgbar=nil)
		return true unless @avira_version.nil?
		@avira_version = Array.new
		scancl = "ruby generic_wrapper.rb /var/run/scancl.version.ret /AntiVir/scancl --version"
		lastupd = Time.now.to_f
		IO.popen(scancl) { |l|
			while l.gets
				line = $_ 
				unless line.strip == "-TickTock-" 
					@avira_version.push(line)
					puts line
				end
				unless pgbar.nil? || Time.now.to_f - lastupd < 0.1
					lastupd = Time.now.to_f
					pgbar.pulse
					pgbar.text = "Prüfe Funktionalität von Avira"
					while (Gtk.events_pending?)
						Gtk.main_iteration
					end
				end
			end
		}
		retval = File.read("/var/run/scancl.version.ret").strip.to_i
		@avira_version = nil if retval > 1
		return false if @avira_version.nil?
	end
	
	# interval: delta in seconds (integer) for the signature update
	# FIXME! Use proxy settings
	# FIXME! Change avupdate script to return proper exit code!
	
	def update_signatures(interval, parameters=[], pgbar=nil)
		return false, "Schlüsseldatei fehlt!" unless File.exist?("/AntiVir/rescue_cd.key")
		return true, nil if File.stat("/AntiVir/rescue_cd.key").mtime.to_f + interval.to_f > Time.now.to_f 
		lastupd = Time.now.to_f
		now = Time.new.strftime("%Y%m%d-%H%M%S") 
		log = File.new(@logdir + "/" + now + "-avupdate.txt", "w") unless @logdir.nil?
		log = File.new(@logdir + "/" + now + "-avupdate.txt", "w") if @logdir.nil? && @debug == true
		command = "ruby generic_wrapper.rb /var/run/avupdate.ret /AntiVirUpdate/avupdate"
		parameters.each { |p| command = command + " '#{p}'" } 
		IO.popen(command) { |l|
			while l.gets
				line = $_
				unless pgbar.nil? || Time.now.to_f - lastupd < 0.1
					lastupd = Time.now.to_f
					pgbar.pulse
					pgbar.text = "Aktualisiere Virensignaturen"
					while (Gtk.events_pending?)
						Gtk.main_iteration
					end
				end
				unless line.strip == "-TickTock-" 
					puts line
					log.write(line) unless @logdir.nil?
				end
			end
		}
		log.close unless @logdir.nil?
		retval = -1
		begin
			retval = File.read("/var/run/avupdate.ret").strip.to_i
		rescue
		end
		puts "avupdate returned: " + retval.to_s
		unless pgbar.nil?
			pgbar.text = "Aktualisierung der Virensignaturen abgeschlossen"
			pgbar.fraction = 1.0
		end
		if retval == 0
			# touch the key file
			system("touch /AntiVir/rescue_cd.key")
			return true, nil
		elsif retval == 1
			# no update available - do nothing
			return true, nil
		else
			# Update failed
			# FIXME! Show dialog informing on the problem
			return false, "Die Aktualisierung der Virensignaturen ist fehlgeschlagen. Wahrscheinlichste Ursache ist eine fehlerhafte Netzwerkkonfiguration."
		end
	end
	
	# FIXME: Use wrapper script
	# FIXME: Use settings!
	# FIXME: Proper handling of boot sector scan
	
	def scan_partition(p, pc=0, pgbar=nil, label=nil)
		if @killed == true
			pgbar.text = "Virenscan abgebrochen" unless pgbar.nil?
			return false
		end
		p.filecount(pgbar)
		p.mount 
		filect = 0
		lastfname = nil
		infected = Array.new
		last_pulse = Time.now.to_f
		unless pgbar.nil?
			pgbar.text = "Partition " + pc.to_s + " von " + @partcount.to_s + " - Starte Scan"
			pgbar.fraction = 0.0 
			while (Gtk.events_pending?)
				Gtk.main_iteration
			end
		end
		# run ruby scancl_wrapper.rb instead 
		# IO.popen("/AntiVir/scancl -a -z -s --showall --log=/dev/null '" + p.mount_point[0] + "'") { |l|
		scancl = "LANGUAGE=en_GB:en LC_ALL=en_GB.UTF-8 ruby scancl_wrapper.rb /var/run/scancl.#{p.uuid}.ret /AntiVir/scancl " + @qparams.join(" ") + " "
		scancl = scancl + "'#{p.mount_point[0]}'" unless @params[-1] =~/^\// 
		puts "Running command: " + scancl
		now = Time.new.strftime("%Y%m%d-%H%M%S") 
		log = File.new(@logdir + "/" + now + "-" + p.device + "-botscan.txt", "w") unless @logdir.nil?
		# For debugging:
		fullog = nil
		if @logdir.nil? && @debug == true
			fulllog = File.new("/tmp/" + now + "-" + p.device + "-botscan-full.txt", "w")
		elsif AntibotMisc.boot_options["debug"] == true
			fulllog = File.new(@logdir + "/" + now + "-" + p.device + "-botscan-full.txt", "w")
		end
		puts "Logging to: " + @logdir + "/" + now + "-" + p.device + "-botscan.txt" unless @logdir.nil?
		prefix = Array.new
		prefix_finished = false
		suffix = Array.new
		suffix_started = false
		file_lines = Array.new
		alert = false
		last_count_change = Time.now.to_f
		last_count = 0
		IO.popen(scancl) { |l|
			while l.gets
				now = Time.now.to_f
				line = $_
				unless line.strip =~ /^-TickTock-$/ || fulllog.nil?
					fulllog.write line
					fulllog.flush
				end
				if prefix_finished == false
					if line.strip =~ /^\//
						prefix_finished = true
						unless @logdir.nil?
							prefix.each { |l| log.write(l) }
							0.upto(3) { |n| log.write("\n") }
						end
					else
						prefix.push(line) unless line.strip =~ /^-TickTock-$/
					end
				end
				
				if line.strip =~ /^Statistics/ || suffix_started == true
					if line.strip =~ /^Statistics/ && alert == true
						# Flush remaining alert
						file_lines.each { |l| log.write(l)  unless @logdir.nil? }
						alert = false
					end
					if line.strip =~ /^Statistics/
						0.upto(3) { |n| log.write("\n") unless @logdir.nil? }
					end
					unless line.strip =~ /^-TickTock-$/
						log.write(line) unless @logdir.nil?
						suffix.push(line)
					end
					suffix_started = true
				end
				
				if line.strip =~ /^\//
					lastfname = line.strip 
					filect += 1
					if file_lines.size > 0 && alert == true
						file_lines.each { |l| log.write(l) unless @logdir.nil? }
						alert = false
					end
					file_lines = Array.new
					file_lines.push(line)
				elsif (line.strip =~ /^-TickTock-$/ ).nil? 
					if line.strip =~ /^ALERT/
						alert = true 
						# Fileout
						fileout = nil
						fileout = ` file -b '#{lastfname}' `.strip if system("which file")
						infected.push AntibotInfectedFile.new(lastfname, p.mount_point[0], p.uuid, line, @qparams, fileout, @logdb, @tstamp)
						puts line.strip 
					end
					file_lines.push(line) unless suffix_started == true || prefix_finished == false 
				end
				
				unless pgbar.nil? || now - last_pulse < 0.1 
					last_pulse = now
					# last_count = filect.to_i
					unless p.filecount(pgbar).nil?
						if filect.to_i > 0
							if filect.to_i > last_count
								last_count = filect.to_i
								last_count_change = now
								pgbar.fraction = filect.to_f / p.filecount(pgbar).to_f
								pgbar.text = "Partition " + pc.to_s + " von " + @partcount.to_s + " - " + filect.to_s + " Dateien gescannt"
							elsif now - last_count_change > 3.0
								pgbar.pulse
								chopped_fname = lastfname.to_s.split("/")[-1] 
								pgbar.text = "Scanne Datei #{chopped_fname}"
							end
						else 
							pgbar.pulse
						end
					end
					while (Gtk.events_pending?)
						Gtk.main_iteration
					end
				end
			end
		}
		log.close unless @logdir.nil?
		unless pgbar.nil?
			pgbar.text = "Partition " + pc.to_s + " von " + @partcount.to_s + " - Scan abgeschlossen"
			pgbar.fraction = 1.0 
			while (Gtk.events_pending?)
				Gtk.main_iteration
			end
		end
		p.umount
		@infections[p] = infected
		@return_codes[p] = File.read("/var/run/scancl.#{p.uuid}.ret").strip.to_i 
		puts "Scanning partition: #{p.device} #{p.uuid} terminated with code #{@return_codes[p]} "
		fulllog.close unless fulllog.nil?
		return false if @killed == true
	end
	
	def run_scan(pgbar=nil, label=nil)
		# Reset variable for kill
		@killed = false
		pc = 1
		failures = Array.new
		( @drivescreen.int_drives + @drivescreen.ext_drives).each { |d|
			d.partitions.each { |p|
				if @drivescreen.saved_states[p.uuid] == true
					scan_partition(p, pc, pgbar, label)
					pc += 1
				end
			}
		}
		if failures.size > 0
			# FIXME: Show error message indicating failed scan!
		end
	end
	
	def dump_scan_results
		
	end
	
end
